@@ -1,5 +1,8 @@
 /*
- * Copyright 1990 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1990 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1996 - 2000, Marek Michałkiewicz
+ * Copyright (c) 2001 - 2006, Tomasz Kłoczko
+ * Copyright (c) 2007 - 2008, Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -10,48 +13,56 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. Neither the name of Julianne F. Haugh nor the names of its contributors
- *    may be used to endorse or promote products derived from this software
- *    without specific prior written permission.
+ * 3. The name of the copyright holders or contributors may not be used to
+ *    endorse or promote products derived from this software without
+ *    specific prior written permission.
  *
- * THIS SOFTWARE IS PROVIDED BY JULIE HAUGH AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED. IN NO EVENT SHALL JULIE HAUGH OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include <config.h>
 
-#ident "$Id: newgrp.c 1876 2008-03-07 20:21:15Z nekral-guest $"
+#ident "$Id: newgrp.c 2982 2009-05-22 11:10:02Z nekral-guest $"
 
 #include <errno.h>
 #include <grp.h>
 #include <pwd.h>
 #include <stdio.h>
+#include <assert.h>
 #include "defines.h"
 #include "getdef.h"
 #include "prototypes.h"
+/*@-exitarg@*/
 #include "exitcodes.h"
+
 /*
  * Global variables
  */
+char *Prog;
+
 extern char **newenvp;
 extern char **environ;
 
 #ifdef HAVE_SETGROUPS
 static int ngroups;
-static GETGROUPS_T *grouplist;
+static /*@null@*/ /*@only@*/GETGROUPS_T *grouplist;
 #endif
 
-static char *Prog;
-static int is_newgrp;
+static bool is_newgrp;
+
+#ifdef WITH_AUDIT
+static char audit_buf[80];
+#endif
 
 /* local function prototypes */
 static void usage (void);
@@ -65,21 +76,22 @@ static void syslog_sg (const char *name, const char *group);
  */
 static void usage (void)
 {
-	if (is_newgrp)
-		fputs (_("Usage: newgrp [-] [group]\n"), stderr);
-	else
-		fputs (_("Usage: sg group [[-c] command]\n"), stderr);
+	if (is_newgrp) {
+		(void) fputs (_("Usage: newgrp [-] [group]\n"), stderr);
+	} else {
+		(void) fputs (_("Usage: sg group [[-c] command]\n"), stderr);
+	}
 }
 
 /*
  * find_matching_group - search all groups of a given group id for
  *                       membership of a given username
  */
-static struct group *find_matching_group (const char *name, gid_t gid)
+static /*@null@*/struct group *find_matching_group (const char *name, gid_t gid)
 {
 	struct group *gr;
 	char **look;
-	int notfound = 1;
+	bool notfound = true;
 
 	setgrent ();
 	while ((gr = getgrent ()) != NULL) {
@@ -92,9 +104,13 @@ static struct group *find_matching_group (const char *name, gid_t gid)
 		 * Test for membership of 'name'.
 		 */
 		look = gr->gr_mem;
-		while (*look && (notfound = strcmp (*look++, name)));
-		if (!notfound)
+		while ((NULL != *look) && notfound) {
+			notfound = (strcmp (*look, name) != 0);
+			look++;
+		}
+		if (!notfound) {
 			break;
+		}
 	}
 	endgrent ();
 	return gr;
@@ -111,7 +127,7 @@ static void check_perms (const struct group *grp,
                          struct passwd *pwd,
                          const char *groupname)
 {
-	int needspasswd = 0;
+	bool needspasswd = false;
 	struct spwd *spwd;
 	char *cp;
 	const char *cpasswd;
@@ -125,8 +141,10 @@ static void check_perms (const struct group *grp,
 	 * anyway.
 	 *
 	 */
-	if (grp->gr_gid != pwd->pw_gid && !is_on_list (grp->gr_mem, pwd->pw_name))
-		needspasswd = 1;
+	if (   (grp->gr_gid != pwd->pw_gid)
+	    && !is_on_list (grp->gr_mem, pwd->pw_name)) {
+		needspasswd = true;
+	}
 
 	/*
 	 * If she does not have either a shadowed password, or a regular
@@ -134,11 +152,13 @@ static void check_perms (const struct group *grp,
 	 * group password.
 	 */
 	spwd = xgetspnam (pwd->pw_name);
-	if (NULL != spwd)
+	if (NULL != spwd) {
 		pwd->pw_passwd = spwd->sp_pwdp;
+	}
 
-	if (pwd->pw_passwd[0] == '\0' && grp->gr_passwd[0])
-		needspasswd = 1;
+	if ((pwd->pw_passwd[0] == '\0') && (grp->gr_passwd[0] != '\0')) {
+		needspasswd = true;
+	}
 
 	/*
 	 * Now I see about letting her into the group she requested. If she
@@ -146,14 +166,15 @@ static void check_perms (const struct group *grp,
 	 * the password. Otherwise I ask for a password if she flunked one
 	 * of the tests above.
 	 */
-	if (getuid () != 0 && needspasswd) {
+	if ((getuid () != 0) && needspasswd) {
 		/*
 		 * get the password from her, and set the salt for
 		 * the decryption from the group file.
 		 */
 		cp = getpass (_("Password: "));
-		if (NULL == cp)
+		if (NULL == cp) {
 			goto failure;
+		}
 
 		/*
 		 * encrypt the key she gave us using the salt from the
@@ -167,22 +188,26 @@ static void check_perms (const struct group *grp,
 		    strcmp (cpasswd, grp->gr_passwd) != 0) {
 #ifdef WITH_AUDIT
 			snprintf (audit_buf, sizeof(audit_buf),
-			          "authentication new-gid=%d", grp->gr_gid);
+			          "authentication new-gid=%lu",
+			          (unsigned long) grp->gr_gid);
 			audit_logger (AUDIT_GRP_AUTH, Prog,
-			              audit_buf, NULL, getuid (), 0);
+			              audit_buf, NULL,
+			              (unsigned int) getuid (), 0);
 #endif
 			SYSLOG ((LOG_INFO,
-				 "Invalid password for group `%s' from `%s'",
+				 "Invalid password for group '%s' from '%s'",
 				 groupname, pwd->pw_name));
-			sleep (1);
-			fputs (_("Invalid password.\n"), stderr);
+			(void) sleep (1);
+			(void) fputs (_("Invalid password.\n"), stderr);
 			goto failure;
 		}
 #ifdef WITH_AUDIT
 		snprintf (audit_buf, sizeof(audit_buf),
-		          "authentication new-gid=%d", grp->gr_gid);
+		          "authentication new-gid=%lu",
+		          (unsigned long) grp->gr_gid);
 		audit_logger (AUDIT_GRP_AUTH, Prog,
-		              audit_buf, NULL, getuid (), 1);
+		              audit_buf, NULL,
+		              (unsigned int) getuid (), 1);
 #endif
 	}
 
@@ -198,13 +223,15 @@ failure:
 		snprintf (audit_buf, sizeof(audit_buf),
 		          "changing new-group=%s", groupname);
 		audit_logger (AUDIT_CHGRP_ID, Prog,
-		              audit_buf, NULL, getuid (), 0);
+		              audit_buf, NULL,
+		              (unsigned int) getuid (), 0);
 	} else {
-		audit_logger (AUDIT_CHGRP_ID, Prog, "changing",
-		              NULL, getuid (), 0);
+		audit_logger (AUDIT_CHGRP_ID, Prog,
+		              "changing", NULL,
+		              (unsigned int) getuid (), 0);
 	}
 #endif
-	exit (1);
+	exit (EXIT_FAILURE);
 }
 
 #ifdef USE_SYSLOG
@@ -219,19 +246,23 @@ static void syslog_sg (const char *name, const char *group)
 	const char *loginname = getlogin ();
 	const char *tty = ttyname (0);
 
-	if (loginname != NULL)
+	if (loginname != NULL) {
 		loginname = xstrdup (loginname);
-	if (tty != NULL)
+	}
+	if (tty != NULL) {
 		tty = xstrdup (tty);
+	}
 
-	if (loginname == NULL)
+	if (loginname == NULL) {
 		loginname = "???";
-	if (tty == NULL)
+	}
+	if (tty == NULL) {
 		tty = "???";
-	else if (strncmp (tty, "/dev/", 5) == 0)
+	} else if (strncmp (tty, "/dev/", 5) == 0) {
 		tty += 5;
+	}
 	SYSLOG ((LOG_INFO,
-		 "user `%s' (login `%s' on %s) switched to group `%s'",
+		 "user '%s' (login '%s' on %s) switched to group '%s'",
 		 name, loginname, tty, group));
 #ifdef USE_PAM
 	/*
@@ -253,14 +284,16 @@ static void syslog_sg (const char *name, const char *group)
 	{
 		pid_t child, pid;
 
-		signal (SIGINT, SIG_IGN);
-		signal (SIGQUIT, SIG_IGN);
-		signal (SIGHUP, SIG_IGN);
-		signal (SIGTSTP, SIG_IGN);
-		signal (SIGTTIN, SIG_IGN);
-		signal (SIGTTOU, SIG_IGN);
+		/* Ignore these signals. The signal handlers will later be
+		 * restored to the default handlers. */
+		(void) signal (SIGINT, SIG_IGN);
+		(void) signal (SIGQUIT, SIG_IGN);
+		(void) signal (SIGHUP, SIG_IGN);
+		(void) signal (SIGTSTP, SIG_IGN);
+		(void) signal (SIGTTIN, SIG_IGN);
+		(void) signal (SIGTTOU, SIG_IGN);
 		child = fork ();
-		if (child < 0) {
+		if ((pid_t)-1 == child) {
 			/* error in fork() */
 			fprintf (stderr, _("%s: failure forking: %s\n"),
 				 is_newgrp ? "newgrp" : "sg", strerror (errno));
@@ -269,14 +302,16 @@ static void syslog_sg (const char *name, const char *group)
 				snprintf (audit_buf, sizeof(audit_buf),
 				          "changing new-group=%s", group);
 				audit_logger (AUDIT_CHGRP_ID, Prog,
-				              audit_buf, NULL, getuid (), 0);
+				              audit_buf, NULL,
+				              (unsigned int) getuid (), 0);
 			} else {
-				audit_logger (AUDIT_CHGRP_ID, Prog, "changing",
-				              NULL, getuid (), 0);
+				audit_logger (AUDIT_CHGRP_ID, Prog,
+				              "changing", NULL,
+				              (unsigned int) getuid (), 0);
 			}
 #endif
-			exit (1);
-		} else if (child) {
+			exit (EXIT_FAILURE);
+		} else if (child != 0) {
 			/* parent - wait for child to finish, then log session close */
 			int cst = 0;
 			gid_t gid = getgid();
@@ -285,42 +320,44 @@ static void syslog_sg (const char *name, const char *group)
 			do {
 				errno = 0;
 				pid = waitpid (child, &cst, WUNTRACED);
-				if (pid == child && WIFSTOPPED (cst)) {
+				if ((pid == child) && (WIFSTOPPED (cst) != 0)) {
 					/* stop when child stops */
-					raise (SIGSTOP);
+					kill (getpid (), WSTOPSIG(cst));
 					/* wake child when resumed */
 					kill (child, SIGCONT);
 				}
-			} while ((pid == child && WIFSTOPPED (cst)) ||
-				 (pid != child && errno == EINTR));
+			} while (   ((pid == child) && (WIFSTOPPED (cst) != 0))
+			         || ((pid != child) && (errno == EINTR)));
 			/* local, no need for xgetgrgid */
 			if (NULL != grp) {
 				SYSLOG ((LOG_INFO,
-				         "user `%s' (login `%s' on %s) returned to group `%s'",
+				         "user '%s' (login '%s' on %s) returned to group '%s'",
 				         name, loginname, tty, grp->gr_name));
 			} else {
 				SYSLOG ((LOG_INFO,
-				         "user `%s' (login `%s' on %s) returned to group `%d'",
-				         name, loginname, tty, gid));
+				         "user '%s' (login '%s' on %s) returned to group '%lu'",
+				         name, loginname, tty,
+				         (unsigned long) gid));
 				/* Either the user's passwd entry has a
 				 * GID that does not match with any group,
 				 * or the group was deleted while the user
 				 * was in a newgrp session.*/
 				SYSLOG ((LOG_WARN,
-				         "unknown GID `%u' used by user `%s'",
-				         gid, name));
+				         "unknown GID '%lu' used by user '%s'",
+				         (unsigned long) gid, name));
 			}
 			closelog ();
-			exit (0);
+			exit ((0 != WIFEXITED (cst)) ? WEXITSTATUS (cst)
+			                             : WTERMSIG (cst) + 128);
 		}
 
 		/* child - restore signals to their default state */
-		signal (SIGINT, SIG_DFL);
-		signal (SIGQUIT, SIG_DFL);
-		signal (SIGHUP, SIG_DFL);
-		signal (SIGTSTP, SIG_DFL);
-		signal (SIGTTIN, SIG_DFL);
-		signal (SIGTTOU, SIG_DFL);
+		(void) signal (SIGINT, SIG_DFL);
+		(void) signal (SIGQUIT, SIG_DFL);
+		(void) signal (SIGHUP, SIG_DFL);
+		(void) signal (SIGTSTP, SIG_DFL);
+		(void) signal (SIGTTIN, SIG_DFL);
+		(void) signal (SIGTTOU, SIG_DFL);
 	}
 #endif				/* USE_PAM */
 }
@@ -331,9 +368,9 @@ static void syslog_sg (const char *name, const char *group)
  */
 int main (int argc, char **argv)
 {
-	int initflag = 0;
+	bool initflag = false;
 	int i;
-	int cflag = 0;
+	bool cflag = false;
 	int err = 0;
 	gid_t gid;
 	char *cp;
@@ -342,20 +379,18 @@ int main (int argc, char **argv)
 	char *command = NULL;
 	char **envp = environ;
 	struct passwd *pwd;
-	struct group *grp;
+	/*@null@*/struct group *grp;
 
 #ifdef SHADOWGRP
 	struct sgrp *sgrp;
 #endif
 
 #ifdef WITH_AUDIT
-	char audit_buf[80];
-
 	audit_help_open ();
 #endif
-	setlocale (LC_ALL, "");
-	bindtextdomain (PACKAGE, LOCALEDIR);
-	textdomain (PACKAGE);
+	(void) setlocale (LC_ALL, "");
+	(void) bindtextdomain (PACKAGE, LOCALEDIR);
+	(void) textdomain (PACKAGE);
 
 	/*
 	 * Save my name for error messages and save my real gid incase of
@@ -390,15 +425,18 @@ int main (int argc, char **argv)
 	initenv ();
 
 	pwd = get_my_pwent ();
-	if (!pwd) {
-		fprintf (stderr, _("unknown UID: %u\n"), getuid ());
+	if (NULL == pwd) {
+		fprintf (stderr, _("%s: Cannot determine your user name.\n"),
+		         Prog);
 #ifdef WITH_AUDIT
-		audit_logger (AUDIT_CHGRP_ID, Prog, "changing", NULL,
-			      getuid (), 0);
+		audit_logger (AUDIT_CHGRP_ID, Prog,
+		              "changing", NULL,
+		              (unsigned int) getuid (), 0);
 #endif
-		SYSLOG ((LOG_WARN, "unknown UID %u", getuid ()));
+		SYSLOG ((LOG_WARN, "Cannot determine the user name of the caller (UID %lu)",
+		         (unsigned long) getuid ()));
 		closelog ();
-		exit (1);
+		exit (EXIT_FAILURE);
 	}
 	name = pwd->pw_name;
 
@@ -419,24 +457,26 @@ int main (int argc, char **argv)
 	 *      sg [-]
 	 *      sg [-] groupid [[-c command]
 	 */
-	if (argc > 0 && (!strcmp (argv[0], "-") || !strcmp (argv[0], "-l"))) {
+	if (   (argc > 0)
+	    && (   (strcmp (argv[0], "-")  == 0)
+	        || (strcmp (argv[0], "-l") == 0))) {
 		argc--;
 		argv++;
-		initflag = 1;
+		initflag = true;
 	}
 	if (!is_newgrp) {
 		/* 
 		 * Do the command line for everything that is
 		 * not "newgrp".
 		 */
-		if (argc > 0 && argv[0][0] != '-') {
+		if ((argc > 0) && (argv[0][0] != '-')) {
 			group = argv[0];
 			argc--;
 			argv++;
 		} else {
 			usage ();
 			closelog ();
-			exit (1);
+			exit (EXIT_FAILURE);
 		}
 		if (argc > 0) {
 
@@ -445,18 +485,19 @@ int main (int argc, char **argv)
 			 * "sg group -c command" (as in the man page) or
 			 * "sg group command" (as in the usage message).
 			 */
-			if (argc > 1 && strcmp (argv[0], "-c") == 0)
+			if ((argc > 1) && (strcmp (argv[0], "-c") == 0)) {
 				command = argv[1];
-			else
+			} else {
 				command = argv[0];
-			cflag++;
+			}
+			cflag = true;
 		}
 	} else {
 		/*
 		 * Do the command line for "newgrp". It's just making sure
 		 * there aren't any flags and getting the new group name.
 		 */
-		if (argc > 0 && argv[0][0] == '-') {
+		if ((argc > 0) && (argv[0][0] == '-')) {
 			usage ();
 			goto failure;
 		} else if (argv[0] != (char *) 0) {
@@ -471,13 +512,15 @@ int main (int argc, char **argv)
 			 */
 			grp = xgetgrgid (pwd->pw_gid);
 			if (NULL == grp) {
-				fprintf (stderr, _("unknown GID: %lu\n"),
-					 (unsigned long) pwd->pw_gid);
-				SYSLOG ((LOG_CRIT, "unknown GID: %lu",
-					 (unsigned long) pwd->pw_gid));
+				fprintf (stderr,
+				         _("%s: GID '%lu' does not exist\n"),
+				         Prog, (unsigned long) pwd->pw_gid);
+				SYSLOG ((LOG_CRIT, "GID '%lu' does not exist",
+				        (unsigned long) pwd->pw_gid));
 				goto failure;
-			} else
+			} else {
 				group = grp->gr_name;
+			}
 		}
 	}
 
@@ -493,8 +536,9 @@ int main (int argc, char **argv)
 	for (;;) {
 		grouplist = (GETGROUPS_T *) xmalloc (i * sizeof (GETGROUPS_T));
 		ngroups = getgroups (i, grouplist);
-		if (i > ngroups && !(ngroups == -1 && errno == EINVAL))
+		if (i > ngroups && !(ngroups == -1 && errno == EINVAL)) {
 			break;
+		}
 		/* not enough room, so try allocating a larger buffer */
 		free (grouplist);
 		i *= 2;
@@ -506,13 +550,15 @@ int main (int argc, char **argv)
 			snprintf (audit_buf, sizeof(audit_buf),
 			          "changing new-group=%s", group);
 			audit_logger (AUDIT_CHGRP_ID, Prog,
-			              audit_buf, NULL, getuid (), 0);
+			              audit_buf, NULL,
+			              (unsigned int) getuid (), 0);
 		} else {
 			audit_logger (AUDIT_CHGRP_ID, Prog,
-			              "changing", NULL, getuid (), 0);
+			              "changing", NULL,
+			              (unsigned int) getuid (), 0);
 		}
 #endif
-		exit (1);
+		exit (EXIT_FAILURE);
 	}
 #endif				/* HAVE_SETGROUPS */
 
@@ -558,7 +604,7 @@ int main (int argc, char **argv)
 	 */
 	grp = getgrnam (group); /* local, no need for xgetgrnam */
 	if (NULL == grp) {
-		fprintf (stderr, _("unknown group: %s\n"), group);
+		fprintf (stderr, _("%s: group '%s' does not exist\n"), Prog, group);
 		goto failure;
 	}
 
@@ -568,7 +614,7 @@ int main (int argc, char **argv)
 	 * membership of the current user.
 	 */
 	grp = find_matching_group (name, grp->gr_gid);
-	if (!grp) {
+	if (NULL == grp) {
 		/*
 		 * No matching group found. As we already know that
 		 * the group exists, this happens only in the case
@@ -577,6 +623,7 @@ int main (int argc, char **argv)
 		 * Re-read the group entry for further processing.
 		 */
 		grp = xgetgrnam (group);
+		assert (NULL != grp);
 	}
 #ifdef SHADOWGRP
 	sgrp = getsgnam (group);
@@ -611,46 +658,62 @@ int main (int argc, char **argv)
 	 * part.
 	 */
 	for (i = 0; i < ngroups; i++) {
-		if (gid == grouplist[i])
+		if (gid == grouplist[i]) {
 			break;
+		}
 	}
 	if (i == ngroups) {
 		if (ngroups >= sysconf (_SC_NGROUPS_MAX)) {
-			fputs (_("too many groups\n"), stderr);
+			(void) fputs (_("too many groups\n"), stderr);
 		} else {
 			grouplist[ngroups++] = gid;
-			if (setgroups (ngroups, grouplist)) {
+			if (setgroups (ngroups, grouplist) != 0) {
 				perror ("setgroups");
 			}
 		}
 	}
 #endif
 
+	/*
+	 * Close all files before changing the user/group IDs.
+	 *
+	 * The needed structure should have been copied before, or
+	 * permission to read the database will be required.
+	 */
+	endspent ();
+#ifdef	SHADOWGRP
+	endsgent ();
+#endif
+	endpwent ();
+	endgrent ();
+
 	/*
 	 * Set the effective GID to the new group id and the effective UID
 	 * to the real UID. For root, this also sets the real GID to the
 	 * new group id.
 	 */
-	if (setgid (gid)) {
+	if (setgid (gid) != 0) {
 		perror ("setgid");
 #ifdef WITH_AUDIT
 		snprintf (audit_buf, sizeof(audit_buf),
-		          "changing new-gid=%d", gid);
+		          "changing new-gid=%lu", (unsigned long) gid);
 		audit_logger (AUDIT_CHGRP_ID, Prog,
-		              audit_buf, NULL, getuid (), 0);
+		              audit_buf, NULL,
+		              (unsigned int) getuid (), 0);
 #endif
-		exit (1);
+		exit (EXIT_FAILURE);
 	}
 
-	if (setuid (getuid ())) {
+	if (setuid (getuid ()) != 0) {
 		perror ("setuid");
 #ifdef WITH_AUDIT
 		snprintf (audit_buf, sizeof(audit_buf),
-		          "changing new-gid=%d", gid);
+		          "changing new-gid=%lu", (unsigned long) gid);
 		audit_logger (AUDIT_CHGRP_ID, Prog,
-		              audit_buf, NULL, getuid (), 0);
+		              audit_buf, NULL,
+		              (unsigned int) getuid (), 0);
 #endif
-		exit (1);
+		exit (EXIT_FAILURE);
 	}
 
 	/*
@@ -662,12 +725,13 @@ int main (int argc, char **argv)
 		execl ("/bin/sh", "sh", "-c", command, (char *) 0);
 #ifdef WITH_AUDIT
 		snprintf (audit_buf, sizeof(audit_buf),
-		          "changing new-gid=%d", gid);
+		          "changing new-gid=%lu", (unsigned long) gid);
 		audit_logger (AUDIT_CHGRP_ID, Prog,
-		              audit_buf, NULL, getuid (), 0);
+		              audit_buf, NULL,
+		              (unsigned int) getuid (), 0);
 #endif
 		perror ("/bin/sh");
-		exit (errno == ENOENT ? E_CMD_NOTFOUND : E_CMD_NOEXEC);
+		exit ((errno == ENOENT) ? E_CMD_NOTFOUND : E_CMD_NOEXEC);
 	}
 
 	/*
@@ -687,35 +751,31 @@ int main (int argc, char **argv)
 	 * problem, try using $SHELL as a workaround; also please notify me
 	 * at jparmele@wildbear.com -- JWP
 	 */
-	if (!initflag && (cp = getenv ("SHELL")))
+	cp = getenv ("SHELL");
+	if (!initflag && (NULL != cp)) {
 		prog = cp;
-	else if (pwd->pw_shell && pwd->pw_shell[0])
+	} else if ((NULL != pwd->pw_shell) && ('\0' != pwd->pw_shell[0])) {
 		prog = pwd->pw_shell;
-	else
+	} else {
 		prog = "/bin/sh";
+	}
 
 	/*
-	 * Now i try to find the basename of the login shell. This will
+	 * Now I try to find the basename of the login shell. This will
 	 * become argv[0] of the spawned command.
 	 */
 	cp = Basename ((char *) prog);
 
-	endspent ();
-#ifdef	SHADOWGRP
-	endsgent ();
-#endif
-	endpwent ();
-	endgrent ();
-
 	/*
 	 * Switch back to her home directory if i am doing login
 	 * initialization.
 	 */
 	if (initflag) {
-		if (chdir (pwd->pw_dir))
+		if (chdir (pwd->pw_dir) != 0) {
 			perror ("chdir");
+		}
 
-		while (*envp) {
+		while (NULL != *envp) {
 			if (strncmp (*envp, "PATH=", 5) == 0 ||
 			    strncmp (*envp, "HOME=", 5) == 0 ||
 			    strncmp (*envp, "SHELL=", 6) == 0 ||
@@ -725,21 +785,26 @@ int main (int argc, char **argv)
 			envp++;
 		}
 	} else {
-		while (*envp)
-			addenv (*envp++, NULL);
+		while (NULL != *envp) {
+			addenv (*envp, NULL);
+			envp++;
+		}
 	}
 
 #ifdef WITH_AUDIT
-	snprintf (audit_buf, sizeof(audit_buf), "changing new-gid=%d", gid);
-	audit_logger (AUDIT_CHGRP_ID, Prog, audit_buf, NULL, getuid (), 1);
+	snprintf (audit_buf, sizeof(audit_buf), "changing new-gid=%lu",
+	          (unsigned long) gid);
+	audit_logger (AUDIT_CHGRP_ID, Prog,
+	              audit_buf, NULL,
+	              (unsigned int) getuid (), 1);
 #endif
 	/*
 	 * Exec the login shell and go away. We are trying to get back to
 	 * the previous environment which should be the user's login shell.
 	 */
 	err = shell (prog, initflag ? (char *) 0 : cp, newenvp);
-	exit (err == ENOENT ? E_CMD_NOTFOUND : E_CMD_NOEXEC);
-	/* NOTREACHED */
+	exit ((err == ENOENT) ? E_CMD_NOTFOUND : E_CMD_NOEXEC);
+	/* @notreached@ */
       failure:
 
 	/*
@@ -754,15 +819,18 @@ int main (int argc, char **argv)
 	 */
 	closelog ();
 #ifdef WITH_AUDIT
-	if (group) {
+	if (NULL != group) {
 		snprintf (audit_buf, sizeof(audit_buf),
 		          "changing new-group=%s", group);
 		audit_logger (AUDIT_CHGRP_ID, Prog, 
-		              audit_buf, NULL, getuid (), 0);
+		              audit_buf, NULL,
+		              (unsigned int) getuid (), 0);
 	} else {
 		audit_logger (AUDIT_CHGRP_ID, Prog,
-		              "changing", NULL, getuid (), 0);
+		              "changing", NULL,
+		              (unsigned int) getuid (), 0);
 	}
 #endif
-	exit (1);
+	exit (EXIT_FAILURE);
 }
+
