@@ -1,5 +1,9 @@
 /*
- * Copyright 1992 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1992 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1996 - 2000, Marek Michałkiewicz
+ * Copyright (c) 2001       , Michał Moskal
+ * Copyright (c) 2001 - 2006, Tomasz Kłoczko
+ * Copyright (c) 2007 - 2009, Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -10,26 +14,26 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. Neither the name of Julianne F. Haugh nor the names of its contributors
- *    may be used to endorse or promote products derived from this software
- *    without specific prior written permission.
+ * 3. The name of the copyright holders or contributors may not be used to
+ *    endorse or promote products derived from this software without
+ *    specific prior written permission.
  *
- * THIS SOFTWARE IS PROVIDED BY JULIE HAUGH AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED. IN NO EVENT SHALL JULIE HAUGH OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include <config.h>
 
-#ident "$Id: pwck.c 1751 2008-02-03 16:28:03Z nekral-guest $"
+#ident "$Id: pwck.c 2883 2009-05-09 21:20:54Z nekral-guest $"
 
 #include <fcntl.h>
 #include <grp.h>
@@ -47,39 +51,69 @@
 /*
  * Exit codes
  */
-
+/*@-exitarg@*/
 #define	E_OKAY		0
 #define	E_USAGE		1
 #define	E_BADENTRY	2
 #define	E_CANTOPEN	3
 #define	E_CANTLOCK	4
 #define	E_CANTUPDATE	5
+#define	E_CANTSORT	6
 
 /*
  * Global variables
  */
+char *Prog;
 
-static char *Prog;
 static const char *pwd_file = PASSWD_FILE;
-static int use_system_pw_file = 1;
+static bool use_system_pw_file = true;
 static const char *spw_file = SHADOW_FILE;
-static int use_system_spw_file = 1;
+static bool use_system_spw_file = true;
 
-static int is_shadow = 0;
+static bool is_shadow = false;
+
+static bool pw_locked  = false;
+static bool spw_locked = false;
 
 /* Options */
-static int read_only = 0;
-static int sort_mode = 0;
-static int quiet = 0;		/* don't report warnings, only errors */
+static bool read_only = false;
+static bool sort_mode = false;
+static bool quiet = false;		/* don't report warnings, only errors */
 
 /* local function prototypes */
+static void fail_exit (int code);
 static void usage (void);
 static void process_flags (int argc, char **argv);
 static void open_files (void);
-static void close_files (int changed);
-static void check_pw_file (int *errors, int *changed);
-static void check_spw_file (int *errors, int *changed);
+static void close_files (bool changed);
+static void check_pw_file (int *errors, bool *changed);
+static void check_spw_file (int *errors, bool *changed);
 
+/*
+ * fail_exit - do some cleanup and exit with the given error code
+ */
+static void fail_exit (int code)
+{
+	if (spw_locked) {
+		if (spw_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, spw_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", spw_dbname ()));
+			/* continue */
+		}
+	}
+
+	if (pw_locked) {
+		if (pw_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, pw_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", pw_dbname ()));
+			/* continue */
+		}
+	}
+
+	closelog ();
+
+	exit (code);
+}
 /*
  * usage - print syntax message and exit
  */
@@ -106,13 +140,13 @@ static void process_flags (int argc, char **argv)
 		switch (arg) {
 		case 'e':	/* added for Debian shadow-961025-2 compatibility */
 		case 'q':
-			quiet = 1;
+			quiet = true;
 			break;
 		case 'r':
-			read_only = 1;
+			read_only = true;
 			break;
 		case 's':
-			sort_mode = 1;
+			sort_mode = true;
 			break;
 		default:
 			usage ();
@@ -137,14 +171,14 @@ static void process_flags (int argc, char **argv)
 	 */
 	if (optind != argc) {
 		pwd_file = argv[optind];
-		pw_name (pwd_file);
-		use_system_pw_file = 0;
+		pw_setdbname (pwd_file);
+		use_system_pw_file = false;
 	}
 	if ((optind + 2) == argc) {
 		spw_file = argv[optind + 1];
-		spw_name (spw_file);
-		is_shadow = 1;
-		use_system_spw_file = 0;
+		spw_setdbname (spw_file);
+		is_shadow = true;
+		use_system_spw_file = false;
 	} else if (optind == argc) {
 		is_shadow = spw_file_present ();
 	}
@@ -163,22 +197,20 @@ static void open_files (void)
 	 */
 	if (!read_only) {
 		if (pw_lock () == 0) {
-			fprintf (stderr, _("%s: cannot lock file %s\n"),
+			fprintf (stderr,
+			         _("%s: cannot lock %s; try again later.\n"),
 			         Prog, pwd_file);
-			if (use_system_pw_file) {
-				SYSLOG ((LOG_WARN, "cannot lock %s", pwd_file));
-			}
-			closelog ();
-			exit (E_CANTLOCK);
+			fail_exit (E_CANTLOCK);
 		}
-		if (is_shadow && (spw_lock () == 0)) {
-			fprintf (stderr, _("%s: cannot lock file %s\n"),
-			         Prog, spw_file);
-			if (use_system_spw_file) {
-				SYSLOG ((LOG_WARN, "cannot lock %s", spw_file));
+		pw_locked = true;
+		if (is_shadow) {
+			if (spw_lock () == 0) {
+				fprintf (stderr,
+				         _("%s: cannot lock %s; try again later.\n"),
+				         Prog, spw_file);
+				fail_exit (E_CANTLOCK);
 			}
-			closelog ();
-			exit (E_CANTLOCK);
+			spw_locked = true;
 		}
 	}
 
@@ -187,22 +219,20 @@ static void open_files (void)
 	 * otherwise.
 	 */
 	if (pw_open (read_only ? O_RDONLY : O_RDWR) == 0) {
-		fprintf (stderr, _("%s: cannot open file %s\n"),
+		fprintf (stderr, _("%s: cannot open %s\n"),
 		         Prog, pwd_file);
 		if (use_system_pw_file) {
 			SYSLOG ((LOG_WARN, "cannot open %s", pwd_file));
 		}
-		closelog ();
-		exit (E_CANTOPEN);
+		fail_exit (E_CANTOPEN);
 	}
 	if (is_shadow && (spw_open (read_only ? O_RDONLY : O_RDWR) == 0)) {
-		fprintf (stderr, _("%s: cannot open file %s\n"),
+		fprintf (stderr, _("%s: cannot open %s\n"),
 		         Prog, spw_file);
 		if (use_system_spw_file) {
 			SYSLOG ((LOG_WARN, "cannot open %s", spw_file));
 		}
-		closelog ();
-		exit (E_CANTOPEN);
+		fail_exit (E_CANTOPEN);
 	}
 }
 
@@ -213,7 +243,7 @@ static void open_files (void)
  *	changes are committed in the databases. The databases are
  *	unlocked anyway.
  */
-static void close_files (int changed)
+static void close_files (bool changed)
 {
 	/*
 	 * All done. If there were no change we can just abandon any
@@ -221,34 +251,44 @@ static void close_files (int changed)
 	 */
 	if (changed) {
 		if (pw_close () == 0) {
-			fprintf (stderr, _("%s: cannot update file %s\n"),
+			fprintf (stderr, _("%s: failure while writing changes to %s\n"),
 			         Prog, pwd_file);
-			SYSLOG ((LOG_WARN, "cannot update %s", pwd_file));
-			closelog ();
-			exit (E_CANTUPDATE);
+			SYSLOG ((LOG_ERR, "failure while writing changes to %s", pwd_file));
+			fail_exit (E_CANTUPDATE);
 		}
 		if (is_shadow && (spw_close () == 0)) {
-			fprintf (stderr, _("%s: cannot update file %s\n"),
+			fprintf (stderr, _("%s: failure while writing changes to %s\n"),
 			         Prog, spw_file);
-			SYSLOG ((LOG_WARN, "cannot update %s", spw_file));
-			closelog ();
-			exit (E_CANTUPDATE);
+			SYSLOG ((LOG_ERR, "failure while writing changes to %s", spw_file));
+			fail_exit (E_CANTUPDATE);
 		}
 	}
 
 	/*
 	 * Don't be anti-social - unlock the files when you're done.
 	 */
-	if (is_shadow) {
-		spw_unlock ();
+	if (spw_locked) {
+		if (spw_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, spw_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", spw_dbname ()));
+			/* continue */
+		}
 	}
-	(void) pw_unlock ();
+	spw_locked = false;
+	if (pw_locked) {
+		if (pw_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, pw_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", pw_dbname ()));
+			/* continue */
+		}
+	}
+	pw_locked = false;
 }
 
 /*
  * check_pw_file - check the content of the passwd file
  */
-static void check_pw_file (int *errors, int *changed)
+static void check_pw_file (int *errors, bool *changed)
 {
 	struct commonio_entry *pfe, *tpfe;
 	struct passwd *pwd;
@@ -262,7 +302,7 @@ static void check_pw_file (int *errors, int *changed)
 		 * If this is a NIS line, skip it. You can't "know" what NIS
 		 * is going to do without directly asking NIS ...
 		 */
-		if ((pfe->line[0] == '+') || (pfe->line[0] == '-')) {
+		if (('+' == pfe->line[0]) || ('-' == pfe->line[0])) {
 			continue;
 		}
 
@@ -294,9 +334,9 @@ static void check_pw_file (int *errors, int *changed)
 			 * loop to try out the next list element.
 			 */
 		      delete_pw:
-			SYSLOG ((LOG_INFO, "delete passwd line `%s'",
+			SYSLOG ((LOG_INFO, "delete passwd line '%s'",
 			         pfe->line));
-			*changed = 1;
+			*changed = true;
 
 			__pw_del_entry (pfe);
 			continue;
@@ -350,11 +390,19 @@ static void check_pw_file (int *errors, int *changed)
 		/*
 		 * Check for invalid usernames.  --marekm
 		 */
-		if (!check_user_name (pwd->pw_name)) {
+		if (!is_valid_user_name (pwd->pw_name)) {
 			printf (_("invalid user name '%s'\n"), pwd->pw_name);
 			*errors += 1;
 		}
 
+		/*
+		 * Check for invalid user ID.
+		 */
+		if (pwd->pw_uid == (uid_t)-1) {
+			printf (_("invalid user ID '%lu'\n"), (long unsigned int)pwd->pw_uid);
+			*errors += 1;
+		}
+
 		/*
 		 * Make sure the primary group exists
 		 */
@@ -365,8 +413,8 @@ static void check_pw_file (int *errors, int *changed)
 			 * No primary group, just give a warning
 			 */
 
-			printf (_("user %s: no group %u\n"),
-			        pwd->pw_name, pwd->pw_gid);
+			printf (_("user '%s': no group %lu\n"),
+			        pwd->pw_name, (unsigned long) pwd->pw_gid);
 			*errors += 1;
 		}
 
@@ -377,7 +425,7 @@ static void check_pw_file (int *errors, int *changed)
 			/*
 			 * Home directory doesn't exist, give a warning
 			 */
-			printf (_("user %s: directory %s does not exist\n"),
+			printf (_("user '%s': directory '%s' does not exist\n"),
 			        pwd->pw_name, pwd->pw_dir);
 			*errors += 1;
 		}
@@ -392,7 +440,7 @@ static void check_pw_file (int *errors, int *changed)
 			/*
 			 * Login shell doesn't exist, give a warning
 			 */
-			printf (_("user %s: program %s does not exist\n"),
+			printf (_("user '%s': program '%s' does not exist\n"),
 			        pwd->pw_name, pwd->pw_shell);
 			*errors += 1;
 		}
@@ -403,7 +451,7 @@ static void check_pw_file (int *errors, int *changed)
 
 		if (is_shadow) {
 			spw = (struct spwd *) spw_locate (pwd->pw_name);
-			if (spw == NULL) {
+			if (NULL == spw) {
 				printf (_("no matching password file entry in %s\n"),
 				        spw_file);
 				printf (_("add user '%s' in %s? "),
@@ -413,25 +461,30 @@ static void check_pw_file (int *errors, int *changed)
 					struct spwd sp;
 					struct passwd pw;
 
-					sp.sp_namp = pwd->pw_name;
-					sp.sp_pwdp = pwd->pw_passwd;
-					sp.sp_min =
+					sp.sp_namp   = pwd->pw_name;
+					sp.sp_pwdp   = pwd->pw_passwd;
+					sp.sp_min    =
 					    getdef_num ("PASS_MIN_DAYS", -1);
-					sp.sp_max =
+					sp.sp_max    =
 					    getdef_num ("PASS_MAX_DAYS", -1);
-					sp.sp_warn =
+					sp.sp_warn   =
 					    getdef_num ("PASS_WARN_AGE", -1);
-					sp.sp_inact = -1;
+					sp.sp_inact  = -1;
 					sp.sp_expire = -1;
-					sp.sp_flag = -1;
-					sp.sp_lstchg =
-					    time ((time_t *) 0) / (24L * 3600L);
-					*changed = 1;
+					sp.sp_flag   = SHADOW_SP_FLAG_UNSET;
+					sp.sp_lstchg = (long) time ((time_t *) 0) / SCALE;
+					if (0 == sp.sp_lstchg) {
+						/* Better disable aging than
+						 * requiring a password change
+						 */
+						sp.sp_lstchg = -1;
+					}
+					*changed = true;
 
 					if (spw_update (&sp) == 0) {
 						fprintf (stderr,
-						         _("%s: can't update shadow entry for %s\n"),
-						         Prog, sp.sp_namp);
+						         _("%s: failed to prepare the new %s entry '%s'\n"),
+						         Prog, spw_dbname (), sp.sp_namp);
 						exit (E_CANTUPDATE);
 					}
 					/* remove password from /etc/passwd */
@@ -439,11 +492,20 @@ static void check_pw_file (int *errors, int *changed)
 					pw.pw_passwd = SHADOW_PASSWD_STRING;	/* XXX warning: const */
 					if (pw_update (&pw) == 0) {
 						fprintf (stderr,
-						         _("%s: can't update passwd entry for %s\n"),
-						         Prog, pw.pw_name);
+						         _("%s: failed to prepare the new %s entry '%s'\n"),
+						         Prog, pw_dbname (), pw.pw_name);
 						exit (E_CANTUPDATE);
 					}
 				}
+			} else {
+				/* The passwd entry has a shadow counterpart.
+				 * Make sure no passwords are in passwd.
+				 */
+				if (strcmp (pwd->pw_passwd, SHADOW_PASSWD_STRING) != 0) {
+					printf (_("user %s has an entry in %s, but its password field in %s is not set to 'x'\n"),
+					        pwd->pw_name, spw_file, pwd_file);
+					*errors += 1;
+				}
 			}
 		}
 	}
@@ -452,7 +514,7 @@ static void check_pw_file (int *errors, int *changed)
 /*
  * check_spw_file - check the content of the shadowed password file (shadow)
  */
-static void check_spw_file (int *errors, int *changed)
+static void check_spw_file (int *errors, bool *changed)
 {
 	struct commonio_entry *spe, *tspe;
 	struct spwd *spw;
@@ -465,7 +527,7 @@ static void check_spw_file (int *errors, int *changed)
 		 * Do not treat lines which were missing in shadow
 		 * and were added earlier.
 		 */
-		if (spe->line == NULL) {
+		if (NULL == spe->line) {
 			continue;
 		}
 
@@ -473,7 +535,7 @@ static void check_spw_file (int *errors, int *changed)
 		 * If this is a NIS line, skip it. You can't "know" what NIS
 		 * is going to do without directly asking NIS ...
 		 */
-		if ((spe->line[0] == '+') || (spe->line[0] == '-')) {
+		if (('+' == spe->line[0]) || ('-' == spe->line[0])) {
 			continue;
 		}
 
@@ -505,9 +567,9 @@ static void check_spw_file (int *errors, int *changed)
 			 * to try out the next list element.
 			 */
 		      delete_spw:
-			SYSLOG ((LOG_INFO, "delete shadow line `%s'",
+			SYSLOG ((LOG_INFO, "delete shadow line '%s'",
 			         spe->line));
-			*changed = 1;
+			*changed = true;
 
 			__spw_del_entry (spe);
 			continue;
@@ -583,8 +645,8 @@ static void check_spw_file (int *errors, int *changed)
 		/*
 		 * Warn if last password change in the future.  --marekm
 		 */
-		if (!quiet
-		    && (spw->sp_lstchg > time ((time_t *) 0) / SCALE)) {
+		if (   !quiet
+		    && (spw->sp_lstchg > (long) time ((time_t *) 0) / SCALE)) {
 			printf (_("user %s: last password change in the future\n"),
 			        spw->sp_namp);
 			*errors += 1;
@@ -598,16 +660,16 @@ static void check_spw_file (int *errors, int *changed)
 int main (int argc, char **argv)
 {
 	int errors = 0;
-	int changed = 0;
+	bool changed = false;
 
 	/*
 	 * Get my name so that I can use it to report errors.
 	 */
 	Prog = Basename (argv[0]);
 
-	setlocale (LC_ALL, "");
-	bindtextdomain (PACKAGE, LOCALEDIR);
-	textdomain (PACKAGE);
+	(void) setlocale (LC_ALL, "");
+	(void) bindtextdomain (PACKAGE, LOCALEDIR);
+	(void) textdomain (PACKAGE);
 
 	OPENLOG ("pwck");
 
@@ -617,11 +679,21 @@ int main (int argc, char **argv)
 	open_files ();
 
 	if (sort_mode) {
-		pw_sort ();
-		if (is_shadow) {
-			spw_sort ();
+		if (pw_sort () != 0) {
+			fprintf (stderr,
+			         _("%s: cannot sort entries in %s\n"),
+			         Prog, pw_dbname ());
+			fail_exit (E_CANTSORT);
 		}
-		changed = 1;
+		if (is_shadow) {
+			if (spw_sort () != 0) {
+				fprintf (stderr,
+				         _("%s: cannot sort entries in %s\n"),
+				         Prog, spw_dbname ());
+				fail_exit (E_CANTSORT);
+			}
+		}
+		changed = true;
 	} else {
 		check_pw_file (&errors, &changed);
 
@@ -637,13 +709,13 @@ int main (int argc, char **argv)
 	/*
 	 * Tell the user what we did and exit.
 	 */
-	if (errors != 0) {
+	if (0 != errors) {
 		printf (changed ?
 		        _("%s: the files have been updated\n") :
 		        _("%s: no changes\n"), Prog);
 	}
 
 	closelog ();
-	exit (errors ? E_BADENTRY : E_OKAY);
+	return ((0 != errors) ? E_BADENTRY : E_OKAY);
 }
 
