@@ -2,7 +2,11 @@
   vipw, vigr  edit the password or group file
   with -s will edit shadow or gshadow file
  
-  Copyright (C) 1997 Guy Maor <maor@ece.utexas.edu>
+  Copyright (c) 1997       , Guy Maor <maor@ece.utexas.edu>
+  Copyright (c) 1999 - 2000, Marek Michałkiewicz
+  Copyright (c) 2002 - 2006, Tomasz Kłoczko
+  Copyright (c) 2007 - 2008, Nicolas François
+  All rights reserved.
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -16,16 +20,18 @@
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
-  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-
-  */
+  Foundation, Inc., 51 Franklin Street, Fifth Floor,
+  Boston, MA 02110-1301, USA.  */
 
 #include <config.h>
 
-#ident "$Id: vipw.c 1741 2008-01-24 19:50:09Z nekral-guest $"
+#ident "$Id: vipw.c 2851 2009-04-30 21:39:38Z nekral-guest $"
 
 #include <errno.h>
 #include <getopt.h>
+#ifdef WITH_SELINUX                                                            
+#include <selinux/selinux.h>                                                   
+#endif
 #include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -34,31 +40,33 @@
 #include <unistd.h>
 #include <utime.h>
 #include "defines.h"
-#include "exitcodes.h"
 #include "groupio.h"
 #include "nscd.h"
 #include "prototypes.h"
 #include "pwio.h"
 #include "sgroupio.h"
 #include "shadowio.h"
+/*@-exitarg@*/
+#include "exitcodes.h"
 
 #define MSG_WARN_EDIT_OTHER_FILE _( \
 	"You have modified %s.\n"\
 	"You may need to modify %s for consistency.\n"\
-	"Please use the command `%s' to do so.\n")
+	"Please use the command '%s' to do so.\n")
 
 /*
  * Global variables
  */
 static const char *progname, *filename, *fileeditname;
-static int filelocked = 0, createedit = 0;
+static bool filelocked = false;
+static bool createedit = false;
 static int (*unlock) (void);
-static int quiet = 0;
+static bool quiet = false;
 
 /* local function prototypes */
 static void usage (void);
 static int create_backup_file (FILE *, const char *, struct stat *);
-static void vipwexit (const char *, int, int);
+static void vipwexit (const char *msg, int syserr, int ret);
 static void vipwedit (const char *, int (*)(void), int (*)(void));
 
 /*
@@ -66,6 +74,7 @@ static void vipwedit (const char *, int (*)(void), int (*)(void));
  */
 static void usage (void)
 {
+	(void) 
 	fputs (_("Usage: vipw [options]\n"
 	         "\n"
 	         "Options:\n"
@@ -90,31 +99,34 @@ static int create_backup_file (FILE * fp, const char *backup, struct stat *sb)
 
 	mask = umask (077);
 	bkfp = fopen (backup, "w");
-	umask (mask);
-	if (!bkfp)
+	(void) umask (mask);
+	if (NULL == bkfp) {
 		return -1;
+	}
 
 	c = 0;
 	if (fseeko (fp, 0, SEEK_SET) == 0)
 		while ((c = getc (fp)) != EOF) {
-			if (putc (c, bkfp) == EOF)
+			if (putc (c, bkfp) == EOF) {
 				break;
+			}
 		}
-	if (c != EOF || ferror (fp) || fflush (bkfp)) {
+	if ((EOF != c) || (ferror (fp) != 0) || (fflush (bkfp) != 0)) {
 		fclose (bkfp);
 		unlink (backup);
 		return -1;
 	}
-	if (fclose (bkfp)) {
+	if (   (fsync (fileno (bkfp)) != 0)
+	    || (fclose (bkfp) != 0)) {
 		unlink (backup);
 		return -1;
 	}
 
 	ub.actime = sb->st_atime;
 	ub.modtime = sb->st_mtime;
-	if (utime (backup, &ub) ||
-	    chmod (backup, sb->st_mode) ||
-	    chown (backup, sb->st_uid, sb->st_gid)) {
+	if (   (utime (backup, &ub) != 0)
+	    || (chmod (backup, sb->st_mode) != 0)
+	    || (chown (backup, sb->st_uid, sb->st_gid) != 0)) {
 		unlink (backup);
 		return -1;
 	}
@@ -128,17 +140,30 @@ static void vipwexit (const char *msg, int syserr, int ret)
 {
 	int err = errno;
 
-	if (createedit)
-		unlink (fileeditname);
-	if (filelocked)
-		(*unlock) ();
-	if (msg)
+	if (createedit) {
+		if (unlink (fileeditname) != 0) {
+			fprintf (stderr, _("%s: failed to remove %s\n"), progname, fileeditname);
+			/* continue */
+		}
+	}
+	if (filelocked) {
+		if ((*unlock) () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), progname, fileeditname);
+			SYSLOG ((LOG_ERR, "failed to unlock %s", fileeditname));
+			/* continue */
+		}
+	}
+	if (NULL != msg) {
 		fprintf (stderr, "%s: %s", progname, msg);
-	if (syserr)
+	}
+	if (0 != syserr) {
 		fprintf (stderr, ": %s", strerror (err));
-	if (!quiet)
-		fprintf (stdout, _("\n%s: %s is unchanged\n"), progname,
+	}
+	(void) fputs ("\n", stderr);
+	if (!quiet) {
+		fprintf (stdout, _("%s: %s is unchanged\n"), progname,
 			 filename);
+	}
 	exit (ret);
 }
 
@@ -165,30 +190,55 @@ vipwedit (const char *file, int (*file_lock) (void), int (*file_unlock) (void))
 	filename = file;
 	fileeditname = fileedit;
 
-	if (access (file, F_OK))
+	if (access (file, F_OK) != 0) {
 		vipwexit (file, 1, 1);
-	if (!file_lock ())
+	}
+#ifdef WITH_SELINUX
+	/* if SE Linux is enabled then set the context of all new files
+	   to be the context of the file we are editing */
+	if (is_selinux_enabled ()) {
+		security_context_t passwd_context=NULL;
+		int ret = 0;
+		if (getfilecon (file, &passwd_context) < 0) {
+			vipwexit (_("Couldn't get file context"), errno, 1);
+		}
+		ret = setfscreatecon (passwd_context);
+		freecon (passwd_context);
+		if (0 != ret) {
+			vipwexit (_("setfscreatecon () failed"), errno, 1);
+		}
+	}
+#endif
+	if (file_lock () == 0) {
 		vipwexit (_("Couldn't lock file"), errno, 5);
-	filelocked = 1;
+	}
+	filelocked = true;
 
 	/* edited copy has same owners, perm */
-	if (stat (file, &st1))
+	if (stat (file, &st1) != 0) {
 		vipwexit (file, 1, 1);
-	if (!(f = fopen (file, "r")))
+	}
+	f = fopen (file, "r");
+	if (NULL == f) {
 		vipwexit (file, 1, 1);
-	if (create_backup_file (f, fileedit, &st1))
+	}
+	if (create_backup_file (f, fileedit, &st1) != 0) {
 		vipwexit (_("Couldn't make backup"), errno, 1);
-	createedit = 1;
+	}
+	createedit = true;
 
 	editor = getenv ("VISUAL");
-	if (!editor)
+	if (NULL == editor) {
 		editor = getenv ("EDITOR");
-	if (!editor)
+	}
+	if (NULL == editor) {
 		editor = DEFAULT_EDITOR;
+	}
 
-	if ((pid = fork ()) == -1)
+	pid = fork ();
+	if (-1 == pid) {
 		vipwexit ("fork", 1, 1);
-	else if (!pid) {
+	} else if (0 == pid) {
 		/* use the system() call to invoke the editor so that it accepts
 		   command line args in the EDITOR and VISUAL environment vars */
 		char *buf;
@@ -200,59 +250,83 @@ vipwedit (const char *file, int (*file_lock) (void), int (*file_unlock) (void))
 			fprintf (stderr, "%s: %s: %s\n", progname, editor,
 				 strerror (errno));
 			exit (1);
-		} else
+		} else {
 			exit (0);
+		}
 	}
 
 	for (;;) {
 		pid = waitpid (pid, &status, WUNTRACED);
-		if (WIFSTOPPED (status)) {
-			kill (getpid (), SIGSTOP);
-			kill (getpid (), SIGCONT);
-		} else
+		if ((pid != -1) && (WIFSTOPPED (status) != 0)) {
+			/* The child (editor) was suspended.
+			 * Suspend vipw. */
+			kill (getpid (), WSTOPSIG(status));
+			/* wake child when resumed */
+			kill (pid, SIGCONT);
+		} else {
 			break;
+		}
 	}
 
-	if (pid == -1 || !WIFEXITED (status) || WEXITSTATUS (status))
+	if (   (-1 == pid)
+	    || (WIFEXITED (status) == 0)
+	    || (WEXITSTATUS (status) != 0)) {
 		vipwexit (editor, 1, 1);
+	}
 
-	if (stat (fileedit, &st2))
+	if (stat (fileedit, &st2) != 0) {
 		vipwexit (fileedit, 1, 1);
-	if (st1.st_mtime == st2.st_mtime)
+	}
+	if (st1.st_mtime == st2.st_mtime) {
 		vipwexit (0, 0, 0);
+	}
+#ifdef WITH_SELINUX                                                            
+	/* unset the fscreatecon */                                             
+	if (is_selinux_enabled ()) {
+		if (setfscreatecon (NULL)) {
+			vipwexit (_("setfscreatecon () failed"), errno, 1);
+		}
+	}
+#endif
 
 	/*
 	 * XXX - here we should check fileedit for errors; if there are any,
 	 * ask the user what to do (edit again, save changes anyway, or quit
 	 * without saving). Use pwck or grpck to do the check.  --marekm
 	 */
-	createedit = 0;
+	createedit = false;
 	unlink (filebackup);
 	link (file, filebackup);
 	if (rename (fileedit, file) == -1) {
 		fprintf (stderr,
-			 _
-			 ("%s: can't restore %s: %s (your changes are in %s)\n"),
-			 progname, file, strerror (errno), fileedit);
+		         _("%s: can't restore %s: %s (your changes are in %s)\n"),
+		         progname, file, strerror (errno), fileedit);
 		vipwexit (0, 0, 1);
 	}
 
-	(*file_unlock) ();
+	if ((*file_unlock) () == 0) {
+		fprintf (stderr, _("%s: failed to unlock %s\n"), progname, fileeditname);
+		SYSLOG ((LOG_ERR, "failed to unlock %s", fileeditname));
+		/* continue */
+	}
+	SYSLOG ((LOG_INFO, "file %s edited", fileeditname));
 }
 
 int main (int argc, char **argv)
 {
-	int editshadow = 0;
+	bool editshadow = false;
 	char *a;
-	int do_vipw;
+	bool do_vipw;
 
-	setlocale (LC_ALL, "");
-	bindtextdomain (PACKAGE, LOCALEDIR);
-	textdomain (PACKAGE);
+	(void) setlocale (LC_ALL, "");
+	(void) bindtextdomain (PACKAGE, LOCALEDIR);
+	(void) textdomain (PACKAGE);
 
 	progname = ((a = strrchr (*argv, '/')) ? a + 1 : *argv);
 	do_vipw = (strcmp (progname, "vigr") != 0);
 
+	OPENLOG (do_vipw ? "vipw" : "vigr");
+
 	{
 		/*
 		 * Parse the command line options.
@@ -264,25 +338,26 @@ int main (int argc, char **argv)
 			{"passwd", no_argument, NULL, 'p'},
 			{"quiet", no_argument, NULL, 'q'},
 			{"shadow", no_argument, NULL, 's'},
+			{NULL, 0, NULL, '\0'}
 		};
 		while ((c =
 			getopt_long (argc, argv, "ghpqs",
 				     long_options, NULL)) != -1) {
 			switch (c) {
 			case 'g':
-				do_vipw = 0;
+				do_vipw = false;
 				break;
 			case 'h':
 				usage ();
 				break;
 			case 'p':
-				do_vipw = 1;
+				do_vipw = true;
 				break;
 			case 'q':
-				quiet = 1;
+				quiet = true;
 				break;
 			case 's':
-				editshadow = 1;
+				editshadow = true;
 				break;
 			default:
 				usage ();
@@ -331,5 +406,6 @@ int main (int argc, char **argv)
 	nscd_flush_cache ("passwd");
 	nscd_flush_cache ("group");
 
-	exit (E_SUCCESS);
+	return E_SUCCESS;
 }
+
